#include <stdio.h>
#include <stdlib.h>

#include "automate_aff_debug.h"
#include "automate_propriete.h"

static int validation_accessible(Automate automate, Etat etat)
{
	return Ensemble_trouver(automate->etats_init, etat);
}

static int validation_coaccessible(Automate automate, Etat etat)
{
	return Ensemble_trouver(automate->etats_finaux, etat);
}

static Ensemble etat_suivants(Etat etat)
{
	if(etat == NULL)
		return NULL;
	return etat->suivants;
}

static Ensemble etat_precedents(Etat etat)
{
	if(etat == NULL)
		return NULL;
	return etat->precedents;
}

static Etat transition_depart(Transition transition)
{
	if(transition == NULL)
		return NULL;
	return transition->depart;
}

static Etat transition_arrivee(Transition transition)
{
	if(transition == NULL)
		return NULL;
	return transition->arrivee;
}

static int cherche_validation(Automate automate, Etat etat, Ensemble faits, int (*validation)(Automate, Etat), Etat (*choix_etat)(Transition), Ensemble (*choix_transitions)(Etat));

static int cherche_validation_dans_transitions(Automate automate, Ensemble faits, AVL cur, int (*validation)(Automate, Etat), Etat (*choix_etat)(Transition), Ensemble (*choix_transitions)(Etat))
{
	if(cur == NULL)
		return 0;
	if(cherche_validation(automate, (*choix_etat)((Transition)(cur->valeur)), faits, validation, choix_etat, choix_transitions))
		return 1;
	if(cherche_validation_dans_transitions(automate, faits, cur->gauche, validation, choix_etat, choix_transitions))
		return 1;
	if(cherche_validation_dans_transitions(automate, faits, cur->droite, validation, choix_etat, choix_transitions))
		return 1;
	return 0;
}

static int cherche_validation(Automate automate, Etat etat, Ensemble faits, int (*validation)(Automate, Etat), Etat (*choix_etat)(Transition), Ensemble (*choix_transitions)(Etat))
{
	if((*validation)(automate, etat))
		return 1;
	if(Ensemble_trouver(faits, etat))
		return 0;
	Ensemble_ajouter(faits, etat);
	if(cherche_validation_dans_transitions(automate, faits, ((*choix_transitions)(etat))->elements, validation, choix_etat, choix_transitions))
		return 1;
	return 0;
}

static int Automate_etat_par_validation(Automate automate, Mot etat_nom, int (*validation)(Automate, Etat), Etat (*choix_etat)(Transition), Ensemble (*choix_transitions)(Etat))
{
	Ensemble faits = NULL;
	Etat etat = NULL;
	int retour = 0;
	
	if(automate == NULL || etat_nom == NULL || validation == NULL || choix_etat == NULL)
		return 0;
	if((etat = Ensemble_trouver_etat_nom(automate->etats, etat_nom)) == NULL)
		return 0;
	if((faits = Ensemble_creer(TYPE_ETAT, 1)) == NULL)
		return 0;
	
	retour = cherche_validation(automate, etat, faits, validation, choix_etat, choix_transitions);
	
	Ensemble_free(&faits);
	
	return retour;
}

int Automate_etat_est_accessible(Automate automate, Mot etat_nom)
{
	return Automate_etat_par_validation(automate, etat_nom, validation_accessible, transition_depart, etat_precedents);
}

int Automate_etat_est_coaccessible(Automate automate, Mot etat_nom)
{
	return Automate_etat_par_validation(automate, etat_nom, validation_coaccessible, transition_arrivee, etat_suivants);
}

int Automate_etat_est_non_accessible(Automate automate, Mot etat_nom)
{
	return ! Automate_etat_est_accessible(automate, etat_nom);
}

int Automate_etat_est_non_coaccessible(Automate automate, Mot etat_nom)
{
	return ! Automate_etat_est_coaccessible(automate, etat_nom);
}

static int Automate_est_deter_trans(const AVL cur, Mot ancien)
{
	if(cur == NULL)
		return 1;
	if(cmp_mot(((Transition)(cur->valeur))->lettre, ancien) == 0)
		return 0;
	if(! Automate_est_deter_trans(cur->gauche, ((Transition)(cur->valeur))->lettre))
		return 0;
	if(! Automate_est_deter_trans(cur->droite, ((Transition)(cur->valeur))->lettre))
		return 0;
	return 1;
}

static int Automate_est_deter_etat(const AVL cur)
{
	if(cur == NULL)
		return 1;
	if(! Automate_est_deter_trans(((Etat)(cur->valeur))->suivants->elements, NULL))
		return 0;
	if(! Automate_est_deter_etat(cur->gauche))
		return 0;
	if(! Automate_est_deter_etat(cur->droite))
		return 0;
	return 1;
}

int Automate_est_deterministe(Automate automate)
{
	if(automate == NULL)
		return 0;
	if(automate->etats_init->elements == NULL)
		return 0;
	if(automate->etats_init->elements->gauche != NULL || automate->etats_init->elements->droite != NULL)
		return 0;
	return Automate_est_deter_etat(automate->etats->elements);
}

static int Automate_est_complet_trans(const AVL cur, Ensemble lettres)
{
	if(cur == NULL)
		return 1;
	if(! Ensemble_trouver(lettres, ((Transition)(cur->valeur))->lettre))
	{
		if(! Ensemble_ajouter(lettres, ((Transition)(cur->valeur))->lettre))
			return 0;
	}
	if(! Automate_est_complet_trans(cur->gauche, lettres))
		return 0;
	if(! Automate_est_complet_trans(cur->droite, lettres))
		return 0;
	return 1;
}

static int Automate_est_complet_etat(Automate automate, const AVL cur)
{
	Ensemble lettres = NULL;
	Ensemble alpha = NULL;
	int retour = 0;
	
	if(cur == NULL)
		return 1;
	if((lettres = Ensemble_creer(TYPE_MOT, 1)) == NULL)
		return 0;
	if(! Ensemble_copier(&alpha, automate->alphabet, 1))
		return 0;
		
	if(Automate_est_complet_trans(((Etat)(cur->valeur))->suivants->elements, lettres)
	&& Ensemble_moins(alpha, lettres))
	{
		retour = alpha->elements == NULL;
	}
	else
		retour = 0;
		
	Ensemble_free(&alpha);
	Ensemble_free(&lettres);
	
	if(! retour)
		return 0;
	if(! Automate_est_complet_etat(automate, cur->gauche))
		return 0;
	if(! Automate_est_complet_etat(automate, cur->droite))
		return 0;
	return 1;
}

int Automate_est_complet(Automate automate)
{
	if(automate == NULL)
		return 0;
	return Automate_est_complet_etat(automate, automate->etats->elements);
}

/*static int Automate_est_standard_aux(Etat etat, const AVL cur)
{
	if(cur == NULL)
		return 1;
	if(((Transition)(cur->valeur))->arrivee == etat)
		return 0;
	if(! Automate_est_standard_aux(etat, cur->gauche))
		return 0;
	if(! Automate_est_standard_aux(etat, cur->droite))
		return 0;
	return 1;
}*/

int Automate_est_standard(Automate automate)
{
	if(automate == NULL)
		return 0;
	if(Ensemble_nombre_elements(automate->etats_init) != 1)
		return 0;
	return ((Etat)(automate->etats_init->elements->valeur))->precedents->elements == NULL;
	/*return Automate_est_standard_aux((Etat)(automate->etats_init->elements->valeur), ((Etat)(automate->etats_init->elements->valeur))->suivants->elements);*/
}

static int Automate_etats_equivalents_aux(Automate automate, Etat prem, Etat sec, Ensemble prem_faits);

static int Automate_etats_equivalents_transition(Automate automate, const AVL cur, Etat sec, Ensemble prem_faits)
{
	Transition transition_sec = NULL;
	Transition transition_prem = NULL;
	AVL tmp = NULL;
	if(cur == NULL)
		return 1;
		
	transition_prem = (Transition)(cur->valeur);
	if((tmp = AVL_trouver(sec->suivants->elements, transition_prem->lettre, &cmp_transition_et_lettre)) == NULL)
		return 0;
	transition_sec = (Transition)(tmp->valeur);
	
	if(! Automate_etats_equivalents_aux(automate, transition_prem->arrivee, transition_sec->arrivee, prem_faits))
		return 0;
	
	if(! Automate_etats_equivalents_transition(automate, cur->gauche, sec, prem_faits))
		return 0;
	if(! Automate_etats_equivalents_transition(automate, cur->droite, sec, prem_faits))
		return 0;
	return 1;
}

static int Automate_etats_equivalents_aux(Automate automate, Etat prem, Etat sec, Ensemble prem_faits)
{
	Mot couple_nom = NULL;
	if(prem == NULL || sec == NULL)
		return 0;
	if(cmp_etat(prem, sec) == 0)
		return 1;
		
	if((couple_nom = couple_chaine(prem->nom, sec->nom, &Mot_chaine, &Mot_chaine)) == NULL)
		return 0;
	if(Ensemble_trouver(prem_faits, couple_nom))
	{
		Element_free(TYPE_MOT)(&couple_nom);
		return 1;
	}
	
	if((Ensemble_trouver(automate->etats_finaux, prem)) ^ (Ensemble_trouver(automate->etats_finaux, sec))
	|| ! Ensemble_ajouter(prem_faits, couple_nom)
	|| Ensemble_nombre_elements(prem->suivants) != Ensemble_nombre_elements(sec->suivants))
	{
		Element_free(TYPE_MOT)(&couple_nom);
		return 0;
	}
	
	if(! Automate_etats_equivalents_transition(automate, prem->suivants->elements, sec, prem_faits))
		return 0;
	return 1;
}

int Automate_etats_equivalents(Automate automate, Mot prem_etat, Mot sec_etat)
{
	int retour;
	Ensemble faits = NULL;
	if(automate == NULL)
		return 0;
	if(! Automate_est_deterministe(automate)
	|| ! Automate_est_complet(automate))
		return 0;
	if((faits = Ensemble_creer(TYPE_MOT, 0)) == NULL)
		return 0;
	retour = Automate_etats_equivalents_aux(automate, Ensemble_trouver_etat_nom(automate->etats, prem_etat), Ensemble_trouver_etat_nom(automate->etats, sec_etat), faits);
	Ensemble_free(&faits);
	return retour;
}
