#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "elements.h"

void (*Element_Ensemble_free(const Ensemble ens))(void *)
{
	if(ens == NULL || ens->liens)
		return &Void_free;
	else
		return Element_free(ens->type);
}

void (*Element_free(Element_Type type))(void *)
{
	switch(type)
	{
		case TYPE_ETAT:
		return &Etat_free;
		case TYPE_TRANSITION:
		return &Transition_free;
		default:
		return &Scalaire_free;
	}
}

void Void_free(void * elem)
{
	return;
}

void Scalaire_free(void * elem)
{
	if(elem == NULL || *((void **)elem) == NULL)
		return;
	free(*((void **)elem));
	*((void **)elem) = NULL;
}

Mot Mot_creer(Mot chaine)
{
	Mot res = NULL;
	if((res = (Mot)malloc((strlen((char *)chaine) + 1) * sizeof(Mot_aux))) == NULL)
		return NULL;
	strcpy((char *)res, (char *)chaine);
	return res;
}

Etat Etat_creer(Mot nom)
{
	Etat etat = NULL;
	if((etat = (Etat)malloc(sizeof(Etat_aux))) == NULL)
		return NULL;
		
	etat->nom = NULL;
	etat->suivants = NULL;
	etat->precedents = NULL;
		
	if((etat->nom = Mot_creer(nom)) == NULL)
	{
		free(etat);
		return NULL;
	}
	if((etat->suivants = Ensemble_creer(TYPE_TRANSITION, 1)) == NULL)
	{
		free(etat);
		free(etat->nom);
		return NULL;
	}
	if((etat->precedents = Ensemble_creer(TYPE_TRANSITION, 1)) == NULL)
	{
		free(etat);
		free(etat->nom);
		free(etat->suivants);
		return NULL;
	}
	
	return etat;
}

void Etat_free(void * etat)
{
	Etat * elem = (Etat *)etat;
	if(elem == NULL || *elem == NULL)
		return;
	Element_free(TYPE_MOT)(&((*elem)->nom));
	Ensemble_free(&((*elem)->suivants));
	Ensemble_free(&((*elem)->precedents));
	free(*elem);
	*elem = NULL;
}

Transition Transition_creer(Etat depart, Mot lettre, Etat arrivee)
{
	Transition transition = NULL;
	if((transition = (Transition)malloc(sizeof(Transition_aux))) == NULL)
		return NULL;
	
	transition->depart = depart;
	transition->arrivee = arrivee;
	
	if((transition->lettre = Mot_creer(lettre)) == NULL)
	{
		free(transition);
		return NULL;
	}
	
	return transition;
}

void Transition_free(void * transition)
{
	Transition * elem = (Transition *)transition;
	if(elem == NULL || *elem == NULL)
		return;
	Element_free(TYPE_MOT)(&((*elem)->lettre));
	free(*elem);
	*elem = NULL;
}

Ensemble Ensemble_creer(Element_Type type, unsigned int liens)
{
	Ensemble ensemble = NULL;
	if((ensemble = (Ensemble)malloc(sizeof(Ensemble_aux))) == NULL)
		return NULL;
		
	ensemble->type = type;
	ensemble->elements = NULL;
	ensemble->liens = (liens != 0);
	
	return ensemble;
}

void Ensemble_free(Ensemble * ensemble)
{
	if(ensemble == NULL || *ensemble == NULL)
		return;
		
	AVL_free(&((*ensemble)->elements), Element_Ensemble_free(*ensemble));
	
	free(*ensemble);
	*ensemble = NULL;
}

void Ensemble_vider(Ensemble ens)
{
	AVL_free(&(ens->elements), Element_Ensemble_free(ens));
}

int (*cmp_type(Element_Type type))(const void *, const void *)
{
	switch(type)
	{
		case TYPE_ENTIER:
		return &cmp_int;
		case TYPE_MOT:
		return &cmp_mot;
		case TYPE_ETAT:
		return &cmp_etat;
		case TYPE_TRANSITION:
		return &cmp_transition;
		default:
		return &cmp_ptr;
	}
}

int cmp_ptr(const void * a, const void * b)
{
	CMP_NULL_VERIF(a, b);
	
	return (int)a - (int)b;
}

int cmp_int(const void * a, const void * b)
{
	CMP_NULL_VERIF(a, b);
	
	return *((int *)a) - *((int *)b);
}

int cmp_mot(const void * a, const void * b)
{
	CMP_NULL_VERIF(a, b);
	
	return strcmp((char *)a, (char *)b);
}

int cmp_etat(const void * a, const void * b)
{
	int retour;
	CMP_NULL_VERIF(a, b);
	
	retour = cmp_mot((void *)(((Etat)a)->nom), (void *)(((Etat)b)->nom));
	
	if(retour == 0)
		return cmp_ptr(a, b);
	return retour;
}

int cmp_transition(const void * a, const void * b)
{
	int retour;
	
	retour = cmp_mot((void *)(((Transition)a)->lettre), (void *)(((Transition)b)->lettre));
	
	if(retour != 0)
		return retour;
		
	retour = cmp_etat((void *)(((Transition)a)->depart), (void *)(((Transition)b)->depart));
	
	if(retour != 0)
		return retour;
		
	retour = cmp_etat((void *)(((Transition)a)->arrivee), (void *)(((Transition)b)->arrivee));
	
	return retour;
}

int cmp_transition_et_lettre(const void * a, const void * b)
{
	CMP_NULL_VERIF(a, b);
	
	return cmp_mot((void *)(((Transition)a)->lettre), b);
}

int cmp_transition_et_depart(const void * a, const void * b)
{
	if(cmp_etat((void *)(((Transition)a)->depart), b) == 0)
		return 0;
	else
		return SPLIT_VALUE;
}

int cmp_transition_et_arrivee(const void * a, const void * b)
{
	if(cmp_etat((void *)(((Transition)a)->arrivee), b) == 0)
		return 0;
	else
		return SPLIT_VALUE;
}

int cmp_etat_et_nom(const void * a, const void * b)
{
	CMP_NULL_VERIF(a, b);
	return cmp_mot((void *)(((Etat)a)->nom), b);
}

int Ensemble_ajouter(Ensemble ens, const void * valeur)
{
	if(ens == NULL)
		return 0;
	return AVL_ajouter(&(ens->elements), valeur, cmp_type(ens->type));
}

int Ensemble_trouver(Ensemble ens, const void * valeur)
{
	if(ens == NULL)
		return 0;
	return AVL_trouver(ens->elements, valeur, cmp_type(ens->type)) != NULL;
}

int Ensemble_trouver_par(Ensemble ens, const void * valeur, Comparaison cmp)
{
	if(ens == NULL)
		return 0;
	if(cmp != cmp_ptr)
	{
		switch(ens->type)
		{
			case TYPE_ENTIER:
			if(cmp != cmp_int)
				return 0;
			break;
			case TYPE_MOT:
			if(cmp != cmp_mot)
				return 0;
			break;
			case TYPE_ETAT:
			if(cmp != cmp_etat
			&& cmp != cmp_etat_et_nom)
				return 0;
			break;
			case TYPE_TRANSITION:
			if(cmp != cmp_transition
			&& cmp != cmp_transition_et_lettre
			&& cmp != cmp_transition_et_depart
			&& cmp != cmp_transition_et_arrivee)
				return 0;
			break;
		}
	}
	if(cmp == cmp_transition_et_depart
	|| cmp == cmp_transition_et_arrivee)
		return AVL_trouver_split(ens->elements, valeur, cmp) != NULL;
	return AVL_trouver(ens->elements, valeur, cmp) != NULL;
}

int Ensemble_supprimer(Ensemble ens, const void * valeur)
{
	AVL retour = NULL;
	void (*elem_aff)(const void *) = Type_afficher((ens)->type, (ens)->liens);
	if(ens == NULL)
		return 0;
		
	retour = AVL_extraire(&(ens->elements), valeur, cmp_type(ens->type));
	
	if(retour == NULL)
		return 0;
			
	AVL_free(&retour, Element_Ensemble_free(ens));
	return 1;
}

Etat Ensemble_trouver_etat_nom(Ensemble ens, Mot nom)
{
	AVL arbre = NULL;
	
	if(ens == NULL)
		return NULL;
	if(ens->type != TYPE_ETAT)
		return NULL;
		
	arbre = AVL_trouver(ens->elements, (void *)nom, &cmp_etat_et_nom);
	
	if(arbre == NULL)
		return NULL;
	return (Etat)(arbre->valeur);
}

Transition Ensemble_trouver_transition(Ensemble ens, Etat depart, Mot lettre, Etat arrivee)
{
	AVL arbre = NULL;
	Transition valeur = NULL;
	
	if(ens == NULL)
		return NULL;
	if(ens->type != TYPE_TRANSITION)
		return NULL;
	
	if((valeur = Transition_creer(depart, lettre, arrivee)) == NULL)	
		return NULL;
	arbre = AVL_trouver(ens->elements, (void *)valeur, cmp_type(TYPE_TRANSITION));
	Element_free(TYPE_TRANSITION)(&valeur);
	
	if(arbre == NULL)
		return NULL;
	return (Transition)(arbre->valeur);
}

int (*Element_copier(const Ensemble ens))(void **, const void *)
{
	if(ens == NULL)
		return NULL;
	if(ens->liens)
		return &lien_copier;
	switch(ens->type)
	{
		case TYPE_ENTIER:
		return &int_copier;
		case TYPE_MOT:
		return &Mot_copier;
		case TYPE_ETAT:
		return &Etat_copier;
		case TYPE_TRANSITION:
		return &Transition_copier;
	}
	return NULL;
}

int int_copier(void ** dest, const void * source)
{
	void * copie = NULL;
	if(dest == NULL || source == NULL)
		return 0;
	if((copie = (void *)malloc(sizeof(int))) == NULL)
		return 0;
	
	**((int **)dest) = *((int *)source);
	
	return 1;
}

int Mot_copier(void ** dest, const void * source)
{
	void * copie = NULL;
	if(dest == NULL || source == NULL)
		return 0;
	if((copie = (void *)Mot_creer((Mot)source)) == NULL)
		return 0;
	
	*dest = copie;
	return 1;
}

int Etat_copier(void ** dest, const void * source)
{
	void * copie = NULL;
	if(dest == NULL || source == NULL)
		return 0;
	
	if((copie = (void *)Etat_creer(((Etat)source)->nom)) == NULL)
		return 0;
	
	*dest = copie;
	return 1;
}

int Transition_copier(void ** dest, const void * source)
{
	void * copie = NULL;
	if(dest == NULL || source == NULL)
		return 0;
	
	/* La transition fais un lien entre deux etats existants */
	if((copie = (void *)Transition_creer(((Transition)source)->depart, ((Transition)source)->lettre, ((Transition)source)->arrivee)) == NULL)
		return 0;
	
	*dest = copie;
	return 1;
}

int lien_copier(void ** dest, const void * source)
{
	if(dest == NULL)
		return 0;
	
	*dest = (void *)source;
	return 1;
}

int Ensemble_copier(Ensemble * dest, const Ensemble source, unsigned int liens)
{
	Ensemble copie = NULL;
	int (*copier)(void **, const void *) = NULL;
	
	if(dest == NULL || source == NULL)
		return 0;
	
	if((copie = Ensemble_creer(source->type, liens || source->liens)) == NULL)
		return 0;
	
	if((copier = Element_copier(copie)) == NULL)
	{
		Ensemble_free(&copie);
		return 0;
	}
	
	if(! AVL_copier(&(copie->elements), source->elements, copier, Element_Ensemble_free(copie)))
	{
		Ensemble_free(&copie);
		return 0;
	}
	
	*dest = copie;
	return 1;
}

int AVL_copier(AVL * dest, const AVL source, int copier(void **, const void *), void (*free_element)(void *))
{
	AVL copie = NULL;
	
	if(dest == NULL)
		return 0;
	if(source == NULL)
	{
		*dest = NULL;
		return 1;
	}
	
	if((copie = AVL_creer_noeud(NULL)) == NULL)
		return 0;
	copie->bal = source->bal;
	if(! copier(&(copie->valeur), source->valeur))
	{
		AVL_free(&copie, free_element);
		return 0;
	}
	
	if(! AVL_copier(&(copie->gauche), source->gauche, copier, free_element)
	|| ! AVL_copier(&(copie->droite), source->droite, copier, free_element))
	{
		AVL_free(&copie, free_element);
		return 0;
	}
	
	*dest = copie;
	return 1;
}

static int Ensemble_union_aux(Ensemble ens, const AVL cur, int copier(void **, const void *), void (*free_element)(void *))
{
	void * valeur = NULL;
	if(cur == NULL)
		return 1;
	if(! Ensemble_trouver(ens, cur->valeur))
	{
		if(! (*copier)(&valeur, cur->valeur))
			return 0;
		if(! Ensemble_ajouter(ens, valeur))
		{
			(*free_element)(&valeur);
			return 0;
		}
	}
	if(! Ensemble_union_aux(ens, cur->gauche, copier, free_element)
	|| ! Ensemble_union_aux(ens, cur->droite, copier, free_element))
		return 0;
	return 1;
}

int Ensemble_union(Ensemble ens, Ensemble ajout)
{
	if(ens == NULL || ajout == NULL)
		return 0;
	if(ens->type != ajout->type)
		return 0;
	return Ensemble_union_aux(ens, ajout->elements, Element_copier(ens), Element_Ensemble_free(ens));
}

static void Ensemble_moins_aux(Ensemble ens, const AVL cur)
{
	if(cur == NULL)
		return;
	Ensemble_supprimer(ens, cur->valeur);
	Ensemble_moins_aux(ens, cur->gauche);
	Ensemble_moins_aux(ens, cur->droite);
}

int Ensemble_moins(Ensemble ens, Ensemble supp)
{
	if(ens == NULL || supp == NULL)
		return 0;
	if(ens->type != supp->type)
		return 0;
	Ensemble_moins_aux(ens, supp->elements);
	return 1;
}

static int Ensemble_nombre_elements_aux(const AVL abr)
{
	if(abr == NULL)
		return 0;
	return 1 + Ensemble_nombre_elements_aux(abr->gauche) + Ensemble_nombre_elements_aux(abr->droite);
}

int Ensemble_egaux(Ensemble ens, Ensemble autre)
{
	Ensemble premier = NULL;
	int retour;
	
	if(! Ensemble_copier(&premier, ens, 1))
		return 0;
	if(! Ensemble_union(premier, autre))
	{
		Ensemble_free(&premier);
		return 0;
	}
	if(! Ensemble_moins(premier, ens))
	{
		Ensemble_free(&premier);
		return 0;
	}
	retour = (Ensemble_nombre_elements(premier) == 0);
	Ensemble_free(&premier);
	
	return retour;
}

int Ensemble_nombre_elements(const Ensemble ens)
{
	if(ens == NULL)
		return 0;
	return Ensemble_nombre_elements_aux(ens->elements);
}

int Ensemble_restructurer_aux(Ensemble ens, const AVL cur)
{	
	if(cur == NULL)
		return 1;
	if(! Ensemble_ajouter(ens, cur->valeur))
		return 0;
	if(! Ensemble_restructurer_aux(ens, cur->gauche))
		return 0;
	if(! Ensemble_restructurer_aux(ens, cur->droite))
		return 0;
	return 1;
}

int Ensemble_restructurer(Ensemble ens)
{
	int retour;
	Ensemble liens = NULL;
	
	if(! Ensemble_copier(&liens, ens, 1))
		return 0;
	AVL_free(&(ens->elements), &Void_free);
	retour = Ensemble_restructurer_aux(ens, liens->elements);
	Ensemble_free(&liens);
	return retour;
}
